/*
 * @COPYRIGHT@
 * 
 * x-kernel v3.3
 * 
 * Copyright (c) 1993,1991,1990,1996  Arizona Board of Regents
 * 
 * @COPYRIGHT@
 *
 * $RCSfile: alpha_kgdb.c,v $
 *
 * HISTORY
 * $Log: alpha_kgdb.c,v $
 * Revision 1.14  1996/02/02  15:54:45  slm
 * Updated copyright and version.
 *
 * Revision 1.13  1995/02/10  00:53:27  davidm
 * (private_memset): new function to avoid having to call memset().
 *
 * Revision 1.12  1995/01/17  20:39:31  davidm
 * Added draina() calls in front of checks to scout_expect_memory_access_fault.
 * This should ensure all outstanding traps are taken before the test
 * executes.
 *
 * Revision 1.11  1994/11/20  01:36:47  davidm
 * Added tracing and FPU support.
 *
 * Revision 1.10  1994/10/30  21:42:59  davidm
 * Added tracing support.
 *
 * Revision 1.9  1994/08/26  01:02:58  davidm
 * Added run-length encoded support (new with GDB 4.13).  Improves
 * response-time significantly!
 *
 * Support for the qOffsets request has been added as well, but GDB 4.13
 * doesn't seem to understand it properly, so it is disabled for now.
 *
 * Revision 1.8  1994/06/11  00:02:38  menze
 * Added copyright notice
 *
 * Revision 1.7  1994/05/14  23:39:13  davidm
 * (kgdb_handle_exception): disable interrupts while in GDB---makes it easier
 * to debug things.
 *
 * Revision 1.6  1994/04/30  22:56:22  davidm
 * P command fixed to use `=' instead of `:'
 *
 * Don't use bcopy() to reduce probability of endless recursion.
 *
 * Revision 1.5  1994/03/29  00:22:05  davidm
 * Various minor fixes.
 *
 * Revision 1.4  1994/03/27  17:03:10  davidm
 * (mem2hex,hex2mem): Fixed such that they work---it is not wise to
 * access memory that may fault by byte---one access may translate into
 * several memory accesses, defeating the "scout_expect_memory_access_fault"
 * mechanism.
 *
 * (regno2addr): New function.
 *
 * (handle_exception): Support for writing individual registers added (P
 * command).  Also, FP is now included in the T response---GDB wants it.
 *
 * (kgdb_handle_exception): Function kgdb_restore_and_rti() is now used to
 * return from an exception reliably.
 *
 * Revision 1.3  1994/03/26  01:55:33  davidm
 * (kgdb_set_debug_traps): no longer invokes kgdb_breakpoint().
 *
 * Revision 1.2  1994/03/17  00:14:42  davidm
 * (handle_exception): added a comment.
 *
 * Revision 1.1  1994/03/16  22:59:06  davidm
 * Initial revision
 *
 */
/*
 * This module provides low-level support for GDB and has been adapted
 * from sparc-stub.c.  Please refer to that file for more on the
 * history of the module.  The contributors seem to be: Hewlett-Packard,
 * Glenn Engel, and Cygnus Support (Stu Grossman).
 *
 * To enable debugger support, two things need to happen.  One, a call
 * to kgdb_set_debug_traps() is necessary in order to allow any
 * breakpoints or error conditions to be properly intercepted and
 * reported to gdb.  Two, a breakpoint needs to be generated to begin
 * communication.  This is most easily accomplished by a call to
 * kgdb_breakpoint().  kgdb_breakpoint() simulates a breakpoint by
 * executing a call_pal bpt.
 *
 *************
 *
 *    The following gdb commands are supported:
 *
 * command	    function				   Return value
 *
 *    g		    return the value of the CPU registers  hex data or ENN
 *    G		    set the value of the CPU registers	   OK or ENN
 *    Pn..=v..	    set CPU register n.. to value v..	   OK or ENN
 *
 *    mAA..AA,LLLL  Read LLLL bytes at address AA..AA	   hex data or ENN
 *    MAA..AA,LLLL: Write LLLL bytes at address AA.AA	   OK or ENN
 *
 *    c		    Resume at current address		   SNN	 ( signal NN)
 *    cAA..AA	    Continue at address AA..AA		   SNN
 *
 *    s		    Step one instruction		   SNN
 *    sAA..AA	    Step one instruction from AA..AA	   SNN
 *
 *    k		    kill
 *
 *    ?		    What was the last sigval ?		   SNN	 (signal NN)
 *
 *   qXXXX	    Request info about XXXX.
 *   QXXXX=yyyy	    Set value of XXXX to yyyy.
 *   qOffsets	    Get section offsets.	     Text=xxx;Data=yyy;Bss=zzz
 *
 *		    Send text to stdout.  Only comes from  Otext
 *		    remote target.
 *
 * All commands and responses are sent with a packet which includes a
 * checksum.  A packet consists of
 *
 * $<packet info>#<checksum>.
 *
 * where
 * <packet info> :: <characters representing the command or response>
 * <checksum>    :: <two hex digits computed as modulo 256 sum of <packetinfo>>
 *
 * When a packet is received, it is first acknowledged with either '+' or '-'.
 * '+' indicates a successful transfer.  '-' indicates a failed transfer.
 *
 * Example:
 *
 * Host:                  Reply:
 * $m0,10#2a               +$00010203040506070809101112131415#42
 *
 * Responses can be run-length encoded to save space.  A '*' means
 * that the next characters is giving a repeat count in the range
 * 3..97.  A repeat count of 3 is encoded as a blank character (0x20)
 * and a repeat count of 97 is encoded as a tilde character (0x7e).
 * The repeat count stands for that many repititions of the character
 * preceding the '*'.  Note that this means that responses cannot
 * contain '*' (this might be a problem in the console-output
 * request).
 *
 * Example: "0* " means the same as "0000".
 */
#include <string.h>
#include <signal.h>
#include <memory.h>
#include <machine/pal.h>
#include "scout.h"
#include "scout_sync_arch.h"
#include "scout_thread.h"
#include "alpha_misc.h"
#include "kgdb.h"

#define initDebugIO	scc_attach
#define putDebugChar	scc_putc
#define getDebugChar	scc_getc

/*
 * external low-level support routines
 */
extern void initDebugIO (void);		/* initialize debugging I/O	 */
extern void putDebugChar (int ch);	/* write a single character      */
extern int  getDebugChar (void);	/* read and return a single char */

/*
 * These defines must match the corresponding code in remote.c.
 */
#define RLE_MIN_COUNT	3
#define RLE_MAX_COUNT	(('~' - ' ') + RLE_MIN_COUNT)

#define BREAKINST	0x00000080	/* call_pal bpt */

/*
 * BUFMAX defines the maximum number of characters in inbound/outbound
 * buffers at least NUMREGBYTES*2 are needed for register packets.
 */
#define BUFMAX 2048

#define NUMREGS		(2*32+1)	/* # of registers */
#define NUMREGBYTES	(NUMREGS * 8)	/* # of bytes for registers */

enum regnames {
      V0,   T0,   T1,   T2,   T3,   T4,   T5,   T6,
      T7,   S0,   S1,   S2,   S3,   S4,   S5,   S6,
      A0,   A1,   A2,   A3,   A4,   A5,   T8,   T9,
     T10,  T11,   RA,  T12,   AT,   GP,   SP, ZERO,
      F0,   F1,   F2,   F3,   F4,   F5,   F6,   F7,
      F8,   F9,  F10,  F11,  F12,  F13,  F14,  F15,
     F16,  F17,  F18,  F19,  F20,  F21,  F22,  F23,
     F24,  F25,  F26,  F27,  F28,  F29,  F30,  F31,
      PC,  VFP
};

#define DBG_STACK		0x001
#define DBG_TRAP		0x002
#define DBG_IN_PACKETS		0x004
#define DBG_OUT_PACKETS		0x008
#define DBG_MEM			0x010
#define DBG_STORE_REG		0x020
#define DBG_CMDS		0x040
#define DBG_SINGLE_STEP		0x080
#define DBG_TRACING		0x100

#ifdef SCOUT_DEBUG
# define DBG(bit,args)	{if (kgdb_debug & (bit)) printf args;}
#else
# define DBG(bit,args)
#endif

int kgdb_debug = 0x100;		/* kgdb debugging level */

static const char hexchars[] = "0123456789abcdef";

static int	nsaved = 0;
static u_int	saved_insn[2];
static u_int	*saved_addr[2];
#ifdef SCOUT_FPU_SUPPORT
 static long	fpu_state[32];
#endif

static int initialized = 0;	/* !0 means we've been initialized */
static char remcomInBuffer[BUFMAX];
static char remcomOutBuffer[BUFMAX];

/*
 * This table contains the mapping between Alpha hardware trap types,
 * and signals, which are primarily what GDB understands.  It also
 * indicates which hardware traps we need to commandeer when
 * initializing the stub.
 */
static const struct hard_trap_info
{
    u_char tt;		/* Trap type code for Alpha */
    u_char signo;	/* Signal that we map this trap into */
} hard_trap_info[] = {
    {0, SIGTRAP},	/* breakpoint instruction */
    {1, SIGILL},	/* illegal instruction (bug check) */
    {2, SIGEMT},	/* gentrap */
    {3, SIGFPE},	/* fpu exception */
    {4, SIGSYS},	/* opDEC fault */
    {0xff, 0xff}	/* must be last */
};

static struct trace_buf {
    void *id;
    u_long *base;
    u_long *volatile limit;
    u_long *next;
} trace_buf = {0, 0, 0, 0};


/*
 * Don't want to use libraries bcopy() to avoid infinite recursion:
 */
static void
private_bcopy(void *from, void *to, int length)
{
    char *fp = from;
    char *tp = to;

    while (length--) *tp++ = *fp++;
} /* private_bcopy */


/*
 * Don't want to use libraries bcopy() to avoid infinite recursion:
 */
static void
private_memset(void *addr, long value, int length)
{
    char *dst = addr;

    while (length--) *dst++ = value;
} /* private_memset */


/*
 * Convert ch from a hex digit to an int:
 */
static int
hex(u_char ch)
{
    if (ch >= 'a' && ch <= 'f') {
	return ch - 'a' + 10;
    } else if (ch >= '0' && ch <= '9') {
	return ch - '0';
    } else if (ch >= 'A' && ch <= 'F') {
	return ch - 'A' + 10;
    } else {
	return -1;
    } /* if */
} /* hex */


/*
 * Scan for the sequence $<data>#<checksum>:
 */
static void
getpacket(char *buffer)
{
    u_char checksum;
    u_char xmitcsum;
    int i;
    int count;
    u_char ch;

    do {
	/* wait around for the start character, ignore all other characters */
	while ((ch = getDebugChar()) != '$') {
	    /* skip */
	} /* while */

	checksum = 0;
	xmitcsum = -1;

	count = 0;

	/* now, read until a # or end of buffer is found */
	while (count < BUFMAX) {
	    ch = getDebugChar();
	    if (ch == '#') {
		break;
	    } /* if */
	    checksum = checksum + ch;
	    buffer[count] = ch;
	    count = count + 1;
	} /* while */

	if (count >= BUFMAX) {
	    continue;
	} /* if */

	buffer[count] = 0;

	if (ch == '#') {
	    xmitcsum = hex(getDebugChar()) << 4;
	    xmitcsum |= hex(getDebugChar());
	    if (checksum != xmitcsum) {
		putDebugChar('-');	/* failed checksum */
	    } else {
		putDebugChar('+'); /* successful transfer */
		/* if a sequence char is present, reply the sequence ID */
		if (buffer[2] == ':') {
		    putDebugChar(buffer[0]);
		    putDebugChar(buffer[1]);
		    /* remove sequence chars from buffer */
		    count = strlen(buffer);
		    for (i=3; i <= count; i++) {
			buffer[i-3] = buffer[i];
		    } /* for */
		} /* if */
	    } /* if */
	} /* if */
    } while (checksum != xmitcsum);

    DBG(DBG_IN_PACKETS, ("kgdb: got packet `$%s'\n", remcomInBuffer));
} /* getpacket */


/*
 * Flush out whatever repeated characters we have accumulated
 * so far.
 */
static u_char*
rle_flush(u_char *dst, u_char rch, int rcount)
{
    if (rcount >= RLE_MIN_COUNT) {
	/*
	 * We have a sequence that's long enough to
	 * warrant encoding, do so:
	 */
	*dst++ = '*';
	*dst++ = (rcount - RLE_MIN_COUNT) + ' ';
    } else {
	/*
	 * The repeated sequence is not long enough to warrant
	 * rle, so output the repeated characters directly:
	 */
	while (rcount--) {
	    *dst++ = rch;
	} /* while */
    } /* if */
    return dst;
} /* rle_flush */


/*
 * Encode sequences of repeated characters by one copy of that
 * character, a star, and a repeat-count encoded as a ASCII character.
 */
static void
rle_encode(u_char *src)
{
    u_char *dst, ch, rch;
    int rcount;

    dst = src;
    rch = '\0';
    rcount = 0;
    while ((ch = *src++) != '\0') {
	if (ch == rch) {
	    ++rcount;
	    if (rcount >= RLE_MAX_COUNT) {
		dst = rle_flush(dst, rch, RLE_MAX_COUNT);
		rcount -= RLE_MAX_COUNT;
	    } /* if */
	} else {
	    dst = rle_flush(dst, rch, rcount);
	    *dst++ = ch;	/* first character is always output */
	    rch = ch;
	    rcount = 0;
	} /* if */
    } /* while */
    dst = rle_flush(dst, rch, rcount);
    *dst++ = '\0';
} /* rle_encode */


/*
 * Send the packet in buffer.
 */
static void
putpacket(u_char *buffer)
{
    u_char checksum;
    int count;
    u_char ch;

    DBG(DBG_OUT_PACKETS, ("kgdb: sending packet `%s'\n", buffer));

    rle_encode(buffer);
    DBG(DBG_OUT_PACKETS, ("kgdb: encoded packet `%s'\n", buffer));

    /*  $<packet info>#<checksum>. */
    do {
	putDebugChar('$');
	checksum = 0;
	count = 0;

	while ((ch = buffer[count]) != '\0') {
	    putDebugChar(ch);
	    checksum += ch;
	    count += 1;
	} /* while */

	putDebugChar('#');
	putDebugChar(hexchars[checksum >> 4]);
	putDebugChar(hexchars[checksum & 0xf]);

	DBG(DBG_OUT_PACKETS, ("kgdb: checksum %#x\n", checksum));
    } while (getDebugChar() == '-');
} /* putpacket */


/*
 * Convert the memory pointed to by mem into hex, placing result in
 * buf.  Return a pointer to the last char put in buf (null), in case
 * of mem fault, return 0.  If MAY_FAULT is non-zero, then we will
 * handle memory faults by returning a 0, else treat a fault like any
 * other fault in the stub.
 */
static u_char *
mem2hex(u_char *mem, u_char *buf, int count, int may_fault)
{
    u_char ch;
    u_long l, *lp;
    int byte;
    
    lp = (u_long*) ((u_long) mem & ~0x7);
    byte = mem - (u_char*) lp;
    scout_expect_memory_access_fault = TRUE;
    l = *lp++;
    l >>= 8 * byte;
    while (count-- > 0) {
	draina();
	if (!scout_expect_memory_access_fault) {
	    return 0;
	} /* if */
	ch = l & 0xff;
	*buf++ = hexchars[ch >> 4];
	*buf++ = hexchars[ch & 0xf];
	l >>= 8;
	if (++byte >= 8) {
	    byte = 0;
	    l = *lp++;
	} /* if */
    } /* while */
    scout_expect_memory_access_fault = FALSE;
    *buf = 0;
    return buf;
} /* mem2hex */


/*
 * Convert the hex array pointed to by buf into binary to be placed in
 * mem return a pointer to the character AFTER the last byte written
 */
static char *
hex2mem(u_char *buf, u_char *mem, int count, int may_fault)
{
    u_char ch;
    u_long l, *lp;
    int byte;
    int i;
    
    lp = (u_long*) ((u_long) mem & ~0x7);
    byte = mem - (u_char*) lp;
    scout_expect_memory_access_fault = TRUE;
    l = *lp;
    for (i = 0; i < count; ++i) {
	draina();
	if (!scout_expect_memory_access_fault) {
	    return 0;
	} /* if */
	ch = hex(*buf++) << 4;
	ch |= hex(*buf++);
	l = (l & ~(0xffL << 8*byte)) | ((long) ch << 8*byte);
	if (++byte >= 8) {
	    byte = 0;
	    *lp++ = l;
	    draina();
	    if (!scout_expect_memory_access_fault) {
		return 0;
	    } /* if */
	    l = *lp;
	} /* if */
    } /* for */
    if (byte) {
	*lp = l;
    } /* if */
    scout_expect_memory_access_fault = FALSE;
    return buf;
} /* hex2mem */


/*
 * Convert the Alpha hardware trap type code to a unix signal number.
 */
static int
compute_signal(int tt)
{
    const struct hard_trap_info *ht;

    for (ht = hard_trap_info; ht->tt < 0xff ; ht++) {
	if (ht->tt == tt) {
	    return ht->signo;
	} /* if */
    } /* for */

    return SIGHUP;		/* default for things we don't know about */
} /* compute_signal */


/*
 * While we find nice hex chars, build an int.
 * Return number of chars processed.
 */
static int
hex2int(char **ptr, int *int_value)
{
    int num_chars = 0;
    int hex_value;

    *int_value = 0;

    while (**ptr) {
	hex_value = hex(**ptr);
	if (hex_value < 0) {
	    break;
	} /* if */

	*int_value = (*int_value << 4) | hex_value;
	num_chars ++;

	(*ptr)++;
    } /* while */

    return num_chars;
} /* hex2int */


/*
 * While we find nice hex chars, build a long.
 * Return number of chars processed.
 */
static int
hex2long(char **ptr, long *long_value)
{
    int num_chars = 0;
    long hex_value;

    *long_value = 0;

    while (**ptr) {
	hex_value = hex(**ptr);
	if (hex_value < 0) {
	    break;
	} /* if */

	*long_value = (*long_value << 4) | hex_value;
	num_chars ++;

	(*ptr)++;
    } /* while */

    return num_chars;
} /* hex2long */


/*
 * Convert a register number to the corresponding pointer into
 * the saved state.  Returns 0 in the case of a bad register.
 */
static u_long*
regno2addr(u_int regno, Alpha_Saved_State *as, u_long **sp)
{
    static u_long zero;
    u_long *addr;

    switch (regno) {
      case V0...S6: addr = &as->v0 + regno; break;
      case A3...AT: addr = &as->a3 + regno - A3; break;
      case A0...A2: addr = &as->a0 + regno - A0; break;
      case SP:	    addr = (u_long*) sp; break;
      case PC:	    addr = (u_long*) &as->pc; break;
      case GP:	    addr = &as->gp; break;
      case ZERO:    addr = &zero; zero = 0L; break;
#ifdef SCOUT_FPU_SUPPORT
      case F0...F31: addr = &fpu_state[regno - F0];
#endif
      default:	    addr = 0; break;
    } /* switch */
    return addr;
} /* regno2addr */


static int
set_breakpoint(Alpha_Saved_State *as, u_long *sp)
{
    int op_code, displ, reg_b, i;
    u_long *reg_b_addr;
    u_int insn;

    scout_expect_memory_access_fault = TRUE;
    insn = *as->pc;
    draina();
    if (!scout_expect_memory_access_fault) {
	return FALSE;
    } /* if */
    scout_expect_memory_access_fault = FALSE;

    op_code = insn >> 26;
    DBG(DBG_SINGLE_STEP, ("kgdb: single-step over "));
    if (op_code >= 0x30) {
	/*
	 * It's a branch: instead of trying to figure out whether the
	 * branch will be taken or not, we'll put a breakpoint at either
	 * location.  This is simpler, more reliable, and probably not
	 * a whole lot slower than the alternative approach of emulating
	 * the branch (emulation can be tricky for fp branches).
	 */
	DBG(DBG_SINGLE_STEP, ("branch"));
	displ = ((int)(insn<<11)) >> 11;
	saved_addr[nsaved++] = as->pc + 1;
	saved_addr[nsaved++] = as->pc + 1 + displ;
    } else if (op_code == 0x1a) {
	/* it's a jump */
	DBG(DBG_SINGLE_STEP, ("jump"));
	reg_b = (insn >> 16) & 0x1f;
	reg_b_addr = regno2addr(reg_b, as, &sp);
	if (reg_b_addr) {
	    saved_addr[nsaved++] = (u_int*) *reg_b_addr;
	} else {
	    return FALSE;
	} /* if */
    } else {
	DBG(DBG_SINGLE_STEP, ("normal"));
	saved_addr[nsaved++] = as->pc + 1;
    } /* if */

    for (i = 0; i < nsaved; ++i) {
	scout_expect_memory_access_fault = TRUE;
	saved_insn[i] = *saved_addr[i];
	draina();
	if (!scout_expect_memory_access_fault) {
	    nsaved = 0;
	    return FALSE;
	} /* if */
	DBG(DBG_SINGLE_STEP, (" -> next_pc=%p", saved_addr[i]));
	*saved_addr[i] = BREAKINST;
	draina();
	if (!scout_expect_memory_access_fault || *saved_addr[i] != BREAKINST) {
	    nsaved = 0;
	    return FALSE;
	} /* if */
	scout_expect_memory_access_fault = FALSE;
    } /* for */
    DBG(DBG_SINGLE_STEP, ("\n"));
    return TRUE;
} /* set_breakpoint */


/*
 * This function does all command processing for interfacing to gdb.  It
 * returns the new stack pointer.
 */
static u_long*
handle_exception(long type, Alpha_Saved_State *as, u_long *sp)
{
    int sigval;
    u_int regno;
    int length;
    char *ptr;
    u_long addr, *reg_addr;

    if (type == 0 && as->pc[-1] != BREAKINST) {
	--as->pc;		/* make pc point to restored insn */
    } /* if */

    DBG(DBG_TRAP, ("kgdb: handle_exception(type=%ld, as=%p, sp=%p, pc=%p)\n",
		   type, as, sp, as->pc));

#if defined(SCOUT_PROFILE) && defined(MCOUNT_TRACE_CALLS)
    {
	extern int tracing_enabled;

	tracing_enabled = 0;
    }
#endif

    /*
     * Reply to host that an exception has occurred.  Include the
     * minimal state GDB needs in order to do anything useful.  The
     * easiest way to find out which registers should be sent is to
     * add a printf to function read_register() in gdb/findvar.c and
     * do an "si" after connecting to the stub.
     */
    sigval = compute_signal(type);
    ptr = remcomOutBuffer;
    *ptr++ = 'T';
    *ptr++ = hexchars[sigval >>  4];
    *ptr++ = hexchars[sigval & 0xf];

    *ptr++ = hexchars[S6 >>  4];	/* s6 also serves as frame pointer */
    *ptr++ = hexchars[S6 & 0xf];
    *ptr++ = ':';
    ptr    = mem2hex((char*)&as->s6, ptr, 8, 0);
    *ptr++ = ';';

    *ptr++ = hexchars[SP >>  4];
    *ptr++ = hexchars[SP & 0xf];
    *ptr++ = ':';
    ptr    = mem2hex((char*)&sp, ptr, 8, 0);
    *ptr++ = ';';

    *ptr++ = hexchars[PC >>  4];
    *ptr++ = hexchars[PC & 0xf];
    *ptr++ = ':';
    ptr    = mem2hex((char*)&as->pc, ptr, 8, 0);
    *ptr++ = ';';

    /*
     * GDB wants VFP --- even though it's a computed value:
     */
    *ptr++ = hexchars[VFP >>  4];
    *ptr++ = hexchars[VFP & 0xf];
    *ptr++ = ':';
    private_memset(ptr, '0', 8 * 2); ptr += 8 * 2;
    *ptr++ = ';';

    *ptr++ = '\0';
    DBG(DBG_OUT_PACKETS, ("kgdb: sending trap info `%s'\n", remcomOutBuffer));
    putpacket(remcomOutBuffer);

    while (1) {
	remcomOutBuffer[0] = '\0';

	getpacket(remcomInBuffer);
	DBG(DBG_CMDS, ("kgdb: got command `%c'\n", remcomInBuffer[0]));
	switch (remcomInBuffer[0]) {
	  case '?':	/* return last signal */
	    remcomOutBuffer[0] = 'S';
	    remcomOutBuffer[1] = hexchars[sigval >> 4];
	    remcomOutBuffer[2] = hexchars[sigval & 0xf];
	    remcomOutBuffer[3] = 0;
	    break;

	  case 'd':	/* toggle debug flag */
	    break;

	  case 'g':	/* return the value of the CPU registers */
	    ptr = remcomOutBuffer;
	    ptr = mem2hex((char*) &as->v0, ptr, 16 * 8, 0); /* v0..s6 */
	    ptr = mem2hex((char*) &as->a0, ptr,  3 * 8, 0); /* a0..a2 */
	    ptr = mem2hex((char*) &as->a3, ptr, 10 * 8, 0); /* a3..at */
	    ptr = mem2hex((char*) &as->gp, ptr,  1 * 8, 0); /* gp */
	    ptr = mem2hex((char*) &sp, ptr,  1 * 8, 0);	    /* sp */
	    private_memset(ptr, '0', 8 * 2);		    /* zero */
	    ptr += 8 * 2;
#ifdef SCOUT_FPU_SUPPORT
	    ptr = mem2hex((char*) fpu_state, ptr, 32 * 8, 0); /* f0..f31 */
#else
	    private_memset(ptr, '0', 32 * 8 * 2);	    /* f0..f31 */
	    ptr += 32 * 8 * 2;
#endif
	    ptr = mem2hex((char*) &as->pc, ptr,  1 * 8, 0); /* pc */
	    break;

	  case 'G':	/* set the value of the CPU registers - return OK */
	    ptr = &remcomInBuffer[1];
	    ptr = hex2mem(ptr, (char*) &as->v0, 16 * 8, 0); /* v0..fp */
	    ptr = hex2mem(ptr, (char*) &as->a0,  3 * 8, 0); /* a0..a2 */
	    ptr = hex2mem(ptr, (char*) &as->a3, 10 * 8, 0); /* a3..at */
	    ptr = hex2mem(ptr, (char*) &as->gp,  1 * 8, 0); /* gp */
	    ptr = hex2mem(ptr, (char*) &sp,  1 * 8, 0); /* sp */
	    ptr += 8 * 2;				    /* zero */
#ifdef SCOUT_FPU_SUPPORT
	    ptr = hex2mem(ptr, (char*) fpu_state, 32 * 8, 0);
#else
	    ptr += 32 * 8 * 2;
#endif
	    ptr = hex2mem(ptr, (char*) &as->pc,  1 * 8, 0); /* pc */
	    strcpy(remcomOutBuffer, "OK");
	    break;

	  case 'P':	/* set the value of one CPU register - return OK */
	    ptr = &remcomInBuffer[1];
	    if (hex2int(&ptr, &regno) && *ptr++ == '=' && regno <= VFP) {
		reg_addr = regno2addr(regno, as, &sp);
		DBG(DBG_STORE_REG, ("kgdb: storing %s to reg %d @ %p "
				    "(as=%p, &sp=%p)\n", ptr, regno, reg_addr,
				    as, &sp));
		if (reg_addr) {
		    hex2mem(ptr, (char*)reg_addr, 8, 0);
		} /* if */
		strcpy(remcomOutBuffer, "OK");
	    } else {
		strcpy(remcomOutBuffer, "E01");
	    } /* if */
	    break;

	  case 'm':	/* mAA..AA,LLLL  Read LLLL bytes at address AA..AA */
	    ptr = &remcomInBuffer[1];
	    if (hex2long(&ptr, &addr) && *ptr++ == ','
		&& hex2int(&ptr, &length))
	    {
		DBG(DBG_MEM, ("kgdb: reading %d bytes at %lx\n",
			      length, addr));
		if (mem2hex((char *)addr, remcomOutBuffer, length, 1)) {
		    break;
		} /* if */
		strcpy(remcomOutBuffer, "E03");
	    } else {
		strcpy(remcomOutBuffer, "E01");
	    } /* if */
	    break;

	  case 'M':
	    /* MAA..AA,LLLL: Write LLLL bytes at address AA.AA return OK */

	    ptr = &remcomInBuffer[1];
	    if (hex2long(&ptr, &addr)
		&& *ptr++ == ','
		&& hex2int(&ptr, &length)
		&& *ptr++ == ':')
	    {
		DBG(DBG_MEM, ("kgdb: writing %d bytes %s at %lx\n", length,
			      ptr, addr));
		if (hex2mem(ptr, (char *)addr, length, 1)) {
		    strcpy(remcomOutBuffer, "OK");
		} else {
		    strcpy(remcomOutBuffer, "E03");
		} /* if */
	    } else {
		strcpy(remcomOutBuffer, "E02");
	    } /* if */
	    break;

	  case 's':	/* sAA..AA    Step at address AA..AA (optional) */
	  case 'c':	/* cAA..AA    Continue at address AA..AA (optional) */
	    /* try to read optional parameter, pc unchanged if no parm: */
	    ptr = &remcomInBuffer[1];
	    if (hex2long(&ptr, &addr)) {
		as->pc = (u_int*) addr;
	    } /* if */
	    DBG(DBG_SINGLE_STEP, ("kgdb: continue/step (%c) at %p\n",
				  remcomInBuffer[0], as->pc));
	    if (remcomInBuffer[0] == 's' && !set_breakpoint(as, sp)) {
		strcpy(remcomOutBuffer, "E03");
		break;
	    } /* if */
	    /*
	     * Need to flush the instruction cache here, as we may
	     * have deposited a breakpoint, and the icache has no way
	     * of knowing that a data ref to some location may have
	     * changed something that is in the instruction cache.
	     */
	    flush_icache();
	    return sp;


	  case 'k':	/* kill the program */
	    break;

	  case 'r':	/* reset */
	    halt();
	    break;

	  case 'q':	/* general query request */
	    if (strcmp (&remcomInBuffer[1], "Offsets") == 0) {
#if 0
		/*
		 * This is how qOffsets is supposed to work.  However,
		 * it just doesn't work.  It maybe because OSF/1 uses
		 * ecoff (see comments in remote.c).
		 *
		 * It doesn't seem to be very useful anyways.
		 */
		extern void _ftext(), _fdata(), _fbss();

		sprintf(remcomOutBuffer, "Text=%lx;Data=%lx;Bss=%lx",
			(long)_ftext/2, (long)_fdata/2, (long)_fbss/2);
#endif
            } else {
		sprintf(remcomOutBuffer,
			"E04, Unknown query %s", &remcomInBuffer[1]);
	    } /* if */
	    break;

	  default:
	    printf("kgdb: handle_exception: unknown command `%c'\n",
		   remcomInBuffer[0]);
	    break;
	} /* switch */

	/* reply to the request */
	putpacket(remcomOutBuffer);
    } /* while */
} /* handle_exception */


#ifdef SCOUT_FPU_SUPPORT

static void
dump_fpu_state(long *to)
{
    asm ("
stt  $f0, 0x00(%0); stt  $f1, 0x08(%0); stt  $f2, 0x10(%0); stt  $f3, 0x18(%0);
stt  $f4, 0x20(%0); stt  $f5, 0x28(%0); stt  $f6, 0x30(%0); stt  $f7, 0x38(%0);
stt  $f8, 0x40(%0); stt  $f9, 0x48(%0); stt $f10, 0x50(%0); stt $f11, 0x58(%0);
stt $f12, 0x60(%0); stt $f13, 0x68(%0); stt $f14, 0x70(%0); stt $f15, 0x78(%0);
stt $f16, 0x80(%0); stt $f17, 0x88(%0); stt $f18, 0x90(%0); stt $f19, 0x98(%0);
stt $f20, 0xa0(%0); stt $f21, 0xa8(%0); stt $f22, 0xb0(%0); stt $f23, 0xb8(%0);
stt $f24, 0xc0(%0); stt $f25, 0xc8(%0); stt $f26, 0xd0(%0); stt $f27, 0xd8(%0);
stt $f28, 0xe0(%0); stt $f29, 0xe8(%0); stt $f30, 0xf0(%0); stt $f31, 0xf8(%0);
	      " :: "r"(to) : "memory");
} /* dump_fpu_state */


restore_fpu_state(long *to)
{
    asm ("
ldt  $f0, 0x00(%0); ldt  $f1, 0x08(%0); ldt  $f2, 0x10(%0); ldt  $f3, 0x18(%0);
ldt  $f4, 0x20(%0); ldt  $f5, 0x28(%0); ldt  $f6, 0x30(%0); ldt  $f7, 0x38(%0);
ldt  $f8, 0x40(%0); ldt  $f9, 0x48(%0); ldt $f10, 0x50(%0); ldt $f11, 0x58(%0);
ldt $f12, 0x60(%0); ldt $f13, 0x68(%0); ldt $f14, 0x70(%0); ldt $f15, 0x78(%0);
ldt $f16, 0x80(%0); ldt $f17, 0x88(%0); ldt $f18, 0x90(%0); ldt $f19, 0x98(%0);
ldt $f20, 0xa0(%0); ldt $f21, 0xa8(%0); ldt $f22, 0xb0(%0); ldt $f23, 0xb8(%0);
ldt $f24, 0xc0(%0); ldt $f25, 0xc8(%0); ldt $f26, 0xd0(%0); ldt $f27, 0xd8(%0);
ldt $f28, 0xe0(%0); ldt $f29, 0xe8(%0); ldt $f30, 0xf0(%0); ldt $f31, 0xf8(%0);
	      " :: "r"(to) : "memory");
} /* restore_fpu_state */

#endif /* SCOUT_FPU_SUPPORT */


/*
 * Set up exception handlers for tracing and breakpoints.
 */
void
kgdb_set_debug_traps()
{
    initialized = 1;
    initDebugIO();
} /* kgdb_set_debug_traps */


/*
 * The following is the nasty part of entering KGDB.  GDB assumes that
 * it can access the memory below the interrupted thread's stack
 * pointer (SP).  This means that the stub cannot run on this stack
 * (interrupts could overwrite changes made by GDB).  Similarly, we do
 * not want to bother the GDB user with the call chain leading from
 * tne interrupt handler to this function.  That is, the SP reported
 * to GDB is the thread's SP at the point it was interrupted.  We
 * achieve all this by copying the saved state and the stack frames
 * below it to a private KGDB stack.  We then switch to that stack;
 * GDB does whatever it wants to the thread's stack; and when it
 * continues execution, we restore the registers from the copy of the
 * saved state and return from the interrupt.
 */
void
kgdb_handle_exception(long type, void *saved_state)
{
    static int break_pending = FALSE;
    static u_long kgdb_stack[1024];
    static u_long *sp = &kgdb_stack[1024];
    register u_long *reg_sp asm ("30");
    u_long *saved_sp, *new_sp;
    Alpha_Saved_State *as;
    int nlongs, i;
    extern void kgdb_restore_and_rti(Alpha_Saved_State *as, u_long *new_sp);

    asm volatile ("" ::: "30");	/* make GCC believe that sp is defined... */

    /*
     * Disable interrupts while in debugger (old IPL will be restored
     * via RTI):
     */
    scout_swap_ipl(SCOUT_MAX_IPL);

    /*
     * Restore original instructions as quickly as possible---in the
     * case the stub executes the same code.
     */
    for (i = 0; i < nsaved; ++i) {
	/* restore original instruction */
	DBG(DBG_SINGLE_STEP, ("kgdb: single-step restoring instruction %x "
			      "at %p\n", saved_insn[i], saved_addr[i]));
	*saved_addr[i] = saved_insn[i];
    } /* for */
    nsaved = 0;

    saved_sp = (u_long*)(saved_state + sizeof(Alpha_Saved_State));

    if (trace_buf.limit && trace_buf.next < trace_buf.limit) {
	static u_int *prev_pc = 0;
	u_int insn, ra, rb, rc;
	u_long *next;
	int access_fault;
	long dist;

	as = (Alpha_Saved_State*) saved_state;
	/*
	 * If the insn causing the breakpoint is not a permanent one,
	 * it has been restored in the above for loop already and we
	 * are going to executed it next:
	 */
	if (as->pc[-1] != BREAKINST) {
	    --as->pc;		/* saved PC is one past breakpoint insn */
	} /* if */

	next = trace_buf.next;

	scout_expect_memory_access_fault = TRUE;
	access_fault = TRUE;	/* assume the worst */
	insn = as->pc[0];	/* get insn we are going to single-step over */
	draina();
	if (scout_expect_memory_access_fault) {
	    scout_expect_memory_access_fault = FALSE;
	    access_fault = FALSE;
	} /* if */

	/*
	 * Encode relative insn to execute next.  If new PC is farther
	 * away than fits into a 32 bit integer, we record it's absolute
	 * address first, with lsb set to 0.  The relative address that
	 * contains the instruction in the top 32 bit has the lsb set.
	 */
	dist = (long)as->pc - (long)prev_pc;
	if (dist < -(1L<<31) || dist >= (1L<<31)) {
	    *next++ = (u_long) as->pc;
	    dist = 0;
	} /* if */
	*next++ = (((u_long) insn)<<32) | ((u_long)dist & 0xffffffff) | 1L;
	prev_pc = as->pc;

	if (access_fault) {
	    *next++ = 0xbadC0FFEEbadF00D;	/* i.e., it's *really* bad */
	    *next++ = 0xbadC0FFEEbadF00D;
	    *next++ = 0xbadC0FFEEbadF00D;
	} else {
	    /* record state that next insn uses (potentially): */
	    rc = insn & 0x1f;
	    rb = (insn >> 16) & 0x1f;
	    ra = (insn >> 21) & 0x1f;
	    *next++ = *regno2addr(ra, as, &saved_sp);
	    *next++ = *regno2addr(rb, as, &saved_sp);
	    *next++ = *regno2addr(rc, as, &saved_sp);
	    /*
	     * Set new breakpoint if at least 5 longs are left in
	     * the trace buffer:
	     */
	    if (next < trace_buf.limit - 5) {
		break_pending = TRUE;
		set_breakpoint(as, saved_sp);
		/* ensure I-stream is coherent with D-stream: */
		flush_icache();
	    } else {
		break_pending = FALSE;	/* no further tracing bpts */
	    } /* if */
	} /* if */
	trace_buf.next = next;
    } else if (break_pending) {
	break_pending = FALSE;	/* simply ignore last tracing breakpoint */
    } else {
	/* copy saved state to exception stack: */
	nlongs = saved_sp - reg_sp;
	as = (Alpha_Saved_State*)
	  (sp - sizeof(Alpha_Saved_State)/sizeof(u_long));
	sp -= nlongs;
	if (sp <= kgdb_stack) {
	    new_sp = sp;
	    sp = kgdb_stack;
	    printf("kgdb: stack underflow (sp=%p, kgdb_stack=%p)!\n",
		   new_sp, kgdb_stack);
	} /* if */
	private_bcopy(reg_sp, sp, nlongs * sizeof(u_long));

	DBG(DBG_STACK, ("kgdb_handle_exception: stack at %p\n", sp));

	/* switch to exception stack: */
	asm volatile ("mov %0,$30" :: "r"(sp) : "memory");

#ifdef SCOUT_FPU_SUPPORT
	dump_fpu_state(fpu_state);
#endif

	new_sp = handle_exception(type, as, saved_sp);

#ifdef SCOUT_FPU_SUPPORT
	restore_fpu_state(fpu_state);
#endif

	/* switch back to process's stack---which might have moved: */
	DBG(DBG_STACK, ("switching stack to %p (saved_sp=%p)\n",
			new_sp, saved_sp));

	sp += nlongs;	/* pop stack */
	kgdb_restore_and_rti(as, new_sp);
    } /* if */
} /* kgdb_handle_exception */


/*
 * This function will generate a breakpoint exception.  It is used at
 * the beginning of a program to sync up with a debugger and can be
 * used otherwise as a quick means to stop program execution and
 * "break" into the debugger.
 */
void
kgdb_breakpoint(void)
{
    if (!initialized) {
	return;
    } /* if */
    asm("call_pal %0" :: "i"(PAL_bpt));
} /* kgdb_breakpoint */


void
kgdb_set_trace_buf(void *id, void *base, u_long len)
{
    u_long *lbase;
#   define ROUND_UP(v,a)	(((v) + (a) - 1) & ~((a) - 1))
#   define ROUND_DOWN(v,a)	((v) & ~((a) - 1))

    lbase = (u_long*) ROUND_UP((u_long)base, sizeof(long));
    len -= (char*) lbase - (char*) base;
    len = ROUND_DOWN(len, 4*sizeof(long));

    DBG(DBG_TRACING,
	("kgdb: trace_buf: id=%p, base=%p, limit=%p (len=%ld)\n",
	 id, lbase, lbase + len / sizeof(long), len));
    trace_buf.id = id;
    trace_buf.base = lbase;
    trace_buf.next = lbase;
    /* important to set this last: */
    trace_buf.limit = lbase + len / sizeof(long);
} /* kgdb_set_trace_buf */


void*
kgdb_stop_tracing(void *id)
{
    if (id == trace_buf.id) {
	trace_buf.limit = 0;	/* atomically stop tracing */

	DBG(DBG_TRACING, ("kgdb: tracing recorded %ld words\n",
			  trace_buf.next - trace_buf.base));
	trace_buf.id = 0;
	return trace_buf.next;
    } /* if */
    return 0;
} /* kgdb_stop_tracing */

			/*** end of kgdb-alpha.c ***/
